#include <stdio.h>
#include "message.h"

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int DdMessageSendDown(DdManager *manager);

int cuddLastLocalIndex(DdManager *manager);

int DdMessageSend2(DdManager *manager,int op, int index, int from, DdNodePar f,DdNodePar g,DdCallTreeNode caller)
{
  message_buffer *buf;
  int i,op_number;
  
  manager->ProcessInfo->WorkDown++;
  manager->ProcessInfo->down_flag=TRUE;

  assert(index<CUDD_MAXINDEX);
  buf=&manager->ProcessInfo->DownBuffer;

  /* find what number the operation has in the current buffer */
  if(from==manager->Mypos)
    {
      op_number=0;
      while(op_number<buf->n_op && buf->ops[op_number]!=op) 
	op_number++;
      if(op_number==DD_MAX_OPS)
	{ /* ran out of space for new operations */
	  DdMessageSendDown(manager);
	  op_number=0;
	}
      if(op_number==buf->n_op)
	{
	  buf->ops[op_number]=op;
	  buf->n_op++;
	}
    }
  else
    {
      op_number=0;
      while(op_number<buf->n_op 
	    && !(buf->ops[op_number]==CUDD_FORWARD_OPERATION 
		 && buf->ForwardInfo[op_number].from==from 
		 && buf->ForwardInfo[op_number].op==op)) 
	op_number++;
      if(op_number==DD_MAX_OPS)
	{ /* ran out of space for new operations */
	  DdMessageSendDown(manager);
	  op_number=0;
	}
      if(op_number==buf->n_op)
	{
	  buf->ops[op_number]=CUDD_FORWARD_OPERATION;
	  buf->ForwardInfo[op_number].from=from; 
	  buf->ForwardInfo[op_number].op=op;
	  buf->n_op++;
	}
    }

  i=buf->n_per_op[op_number];
  buf->f_index [op_number][i]=f.index;
  buf->f_page  [op_number][i]=f.address.page;
  buf->f_number[op_number][i]=f.address.number;
  buf->g_index [op_number][i]=g.index;
  buf->g_page  [op_number][i]=g.address.page;
  buf->g_number[op_number][i]=g.address.number;
  buf->caller  [op_number][i]=caller;

  buf->n++;
  buf->n_per_op[op_number]++;

  if(buf->n>=manager->ProcessInfo->DownCountSend)
    DdMessageSendDown(manager);

  return(1); /* NO ERROR HANDELING */
}

int DdMessageSend3(DdManager *manager,int op, int index, int from, DdNodePar f,DdNodePar g,DdNodePar h,DdCallTreeNode caller)
{
  int info=0;
  assert(index<CUDD_MAXINDEX);

  pvm_setsbuf(manager->ProcessInfo->DownBuf);

  info+=pvm_pkint(&op,1,1);
  info+=pvm_pkint(&index,1,1);  
  info+=pvm_pkint(&from,1,1);
  info+=PVM_pkDdNodePar(f);
  info+=PVM_pkDdNodePar(g);
  info+=PVM_pkDdNodePar(h);
  info+=PVM_pkDdCallTreeNode(caller);

  if(info<0)
    return(0);
  else
    return(DdMessageSendDown(manager));
}


int flush_buffer_down(DdManager *manager)
{
  int type;
  ProcessInfo *pi;
  message_buffer *buf;

  pi=manager->ProcessInfo;
  buf=&pi->DownBuffer;

  assert(manager->Mypos<manager->ProcessInfo->Nprocess-1);
  assert(pi->down_flag==TRUE);
  
  pi->down_flag=FALSE;
  pvm_setsbuf(pi->DownBuf);

  DdPackDownBuffer(buf);

  type=CUDD_FLUSH_DOWN;
  pvm_pkint(&type,1,1);
  pvm_pkint(&pi->flush_count,1,1);

  pvm_send(pi->Tids[manager->Mypos+1],DD_DOWN_MESSAGE);
  pvm_freebuf(pi->DownBuf);
  pi->DownBuf=pvm_mkbuf(pi->Encoding);
  buf->n=0;
  buf->n_op=0;
  pi->DownCountSend=DD_DOWN_BUFFER_SIZE;
#ifdef DD_COMM_MONITOR
  pi->DownMessages++;
#endif
  return(1); /* NO ERROR HANDELING */
}

/*+++++++++++++++++++++++++++++++ reorder stuff ++++++++++++++++++++++++*/

int cuddSendExternRef(DdManager *manager) 
{
  int index,page_number,number;
  DdNodeMore *ref_page;
  int bufid;

  if(manager->Mypos<manager->ProcessInfo->Nprocess-1)
    { /* last process does not send anything */ 
      bufid = pvm_mkbuf(manager->ProcessInfo->Encoding);
      pvm_setsbuf(bufid);
      
      for(index=cuddLastLocalIndex(manager)+1;index<manager->size;index++)
	{
	  pvm_pkint(&index,1,1);
	  for(page_number=0;page_number<manager->memory[index].TableSize;page_number++)
	    {
	      ref_page=manager->memory[index].PageTable2[page_number];
	      if(ref_page!=NULL)
		{ /* hashing should be considered */
		  pvm_pkint(&page_number,1,1);
		  for(number=1;number<=DD_MEM_CHUNK;number++)
		    if(ref_page[number].Reorder.Ref>0)
		      pvm_pkint(&number,1,1);
		  number=NULL_NUMBER;
		  pvm_pkint(&number,1,1);
		}
	    }
	  page_number=NULL_PAGE;
	  pvm_pkint(&page_number,1,1);
	}
      index=CUDD_MAXINDEX;
      pvm_pkint(&index,1,1);
      
      pvm_send(manager->ProcessInfo->Tids[manager->Mypos+1],0);
      pvm_freebuf(bufid);
    }
  return(1); /* NO ERROR HANDELING */
}

int cuddExternGetNewLocations(DdManager *manager,DdMemory *old_mems) 
{
  int index,page_number,i;
  DdNodeMore *more_page;
  DdMemory *mem;
  int extrn;
  int last;

  if(manager->Mypos<manager->ProcessInfo->Nprocess-1)
    { /* last process does not receive anything */ 
      pvm_recv(manager->ProcessInfo->Tids[manager->Mypos+1],0);

      last=cuddLastLocalIndex(manager);
      for(index=last+1;index<manager->size;index++)
	{ 
	  mem=&old_mems[index];
	  for(page_number=0;page_number<mem->TableSize;page_number++)
	    {
	      more_page=mem->PageTable2[page_number];
	      if(more_page!=NULL)
		for(i=1;i<=DD_MEM_CHUNK;i++)
		  if(more_page[i].Reorder.Ref>0)
		    {
		      extrn=more_page[i].Reorder.Ref==DD_EXTERN_REF;
		      PVM_upkDdNodePar(more_page[i].NewLocation);
		      if(extrn) {
			PVM_pkDdNodePar(more_page[i].NewLocation);
		      }
		    }
	    }
	}
    }
  return(1); /* NO ERROR HANDELING */
}

void DdPackDownBuffer(message_buffer *buf)
{
  int n,i;
  for(i=0;i<buf->n_op;i++)
    {
      pvm_pkint(&buf->ops[i],1,1);
      if(buf->ops[i]==CUDD_FORWARD_OPERATION) {
	pvm_pkint(&buf->ForwardInfo[i].from,1,1);
	pvm_pkint(&buf->ForwardInfo[i].op,1,1);
      }
      n=buf->n_per_op[i];
      pvm_pkint(&n,1,1);
      pvm_pkushort(buf->f_index[i], n,1);
      pvm_pkushort(buf->f_page[i],  n,1);
      pvm_pkshort (buf->f_number[i],n,1);
      pvm_pkushort(buf->g_index[i], n,1);
      pvm_pkushort(buf->g_page[i],  n,1);
      pvm_pkshort (buf->g_number[i],n,1);
      pvm_pkint   (buf->caller[i],  n,1);
      buf->n_per_op[i]=0;
    }
}

void DdPackUpBuffer(message_buffer *buf)
{
  int n,i;
  for(i=0;i<buf->n_op;i++)
    {
      pvm_pkint(&buf->ops[i],1,1);
      if(buf->ops[i]==CUDD_FORWARD_REPLY) {
	pvm_pkint(&buf->ForwardInfo[i].to,1,1);
      }
      n=buf->n_per_op[i];
      pvm_pkint(&n,1,1);
      pvm_pkushort(buf->f_index[i], n,1);
      pvm_pkushort(buf->f_page[i],  n,1);
      pvm_pkshort (buf->f_number[i],n,1);
      pvm_pkint   (buf->caller[i],  n,1);
      buf->n_per_op[i]=0;
    }
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/


static int DdMessageSendDown(DdManager *manager)
{
  ProcessInfo *pi;
  message_buffer *buf;
  int stop_op=CUDD_NO_MORE_MESSAGES;

  pi=manager->ProcessInfo;
  buf=&pi->DownBuffer;

  assert(manager->Mypos+1<pi->Nprocess);

  pvm_setsbuf(pi->DownBuf);

  DdPackDownBuffer(buf);
  pvm_pkint(&stop_op,1,1);
  pvm_send(pi->Tids[manager->Mypos+1],DD_DOWN_MESSAGE);

  pvm_freebuf(pi->DownBuf);
  pi->DownBuf=pvm_mkbuf(pi->Encoding);

  buf->n=0;
  buf->n_op=0;
  pi->DownCountSend*=DD_DOWN_BUFFER_GROWTH;
  if(pi->DownCountSend>DD_DOWN_BUFFER_MAX)
    pi->DownCountSend=DD_DOWN_BUFFER_MAX;
#ifdef DD_COMM_MONITOR
  pi->DownMessages++;
#endif

  return(1); /* NO ERROR HANDELING */
}


int cuddLastLocalIndex(DdManager *manager)
{
  int index=0;
  int found=FALSE;
  while(index<manager->size && manager->Local[index]<=manager->Mypos)
    index++;
  return index-1;
}






