#include "message.h"
#include "cuddInt.h" 

#define DdCallTreeCreateAndReturnRes(man,Node,node) DdCallTreeReturnRes(man,Node,node,cuddUniqueInterSafePar(manager,node->index,node->Then.Result,node->Else.Result))

static int DdCallTreeReturnRes(DdManager *manager,DdCallTreeNode Node,
			       DdCallTreeNodeStruct *node,DdNodePar res);
static int DdCallTreeResize(DdManager *manager);
static int DdCallTreeOr(DdManager *manager,DdCallTreeNode Node);

int DdCallTreeCreate(DdManager *manager)
{
  DdCallTree *Tree;
  int i;
  Tree=&manager->CallTree;
  Tree->Size=DD_CALL_TREE_SIZE;
  Tree->NodeList=ALLOC(DdCallTreeNodeStruct,DD_CALL_TREE_SIZE)-1; /* -1 => 1 indexed */
  if(Tree->NodeList==NULL) return(0);
  manager->memused += sizeof(struct DdCallTreeNode) * DD_CALL_TREE_SIZE;
  for(i=1;i<DD_CALL_TREE_SIZE;i++)
    Tree->NodeList[i].Parent=i+1;
  Tree->NodeList[DD_CALL_TREE_SIZE].Parent=NULL_CALL_TREE_NODE;
  Tree->FreeList=1;
  return(1);
}

DdCallTreeNode DdCallTreeCreateNode(DdManager *manager,
				    DdCallTreeNode Parent,
				    int ParentProcess,
				    int index,
				    int type)
{
  DdCallTreeNodeStruct *node;
  DdCallTreeNode Node;

  if(manager->CallTree.FreeList==NULL_CALL_TREE_NODE) {
    if(manager->CallTree.Size>=1<<16)
      while(manager->CallTree.FreeList==NULL_CALL_TREE_NODE)
	DdForceReply(manager);
    else
      if(DdCallTreeResize(manager)==0)
	return(NULL_CALL_TREE_NODE);
  }
  Node=manager->CallTree.FreeList;
  node=&manager->CallTree.NodeList[Node];
  manager->CallTree.FreeList=node->Parent;

  node->Parent=Parent;
  node->ParentProcess=ParentProcess;
  node->index=index;
  node->state=type;
  node->Then.CacheEntry=NULL;
  node->Else.CacheEntry=NULL;
  
  return(Node);
}


int DdCallTreeNodeSetCacheEntry(DdManager *manager,DdCallTreeNode Node,DdCache *entry)
{
  DdCallTreeNodeStruct *node;

  assert(Node!=0);
  node=&manager->CallTree.NodeList[abs(Node)];
  if(Node>0)
    node->Then.CacheEntry=entry;
  else
    node->Else.CacheEntry=entry;
  return(1);
}

/*  

Description 
updates the call tree and proipergates information up if both results
have arrieved at the node.

returns 1 if succesful, 0 if something went wrong

*/

int DdCallTreeUpdateNode(DdManager *manager,DdCallTreeNode Node,DdNodePar Res)
{
  DdCallTreeNodeStruct *node;
  DdCache *CacheEntry=NULL;

  assert(Node!=0);
  node=&manager->CallTree.NodeList[abs(Node)];
  switch(node->state)
    {
    case CUDD_PROPAGATE:
      /* using Then to store CacheEntry */
      CacheEntry=node->Then.CacheEntry;
      DdCallTreeReturnRes(manager,Node,node,Res);
      node->state=CUDD_CALL_TREE_DONE;
      break;

    case CUDD_PROPAGATE_NEGATE:
      /* using Then to store CacheEntry */
      CacheEntry=node->Then.CacheEntry;
      DdCallTreeReturnRes(manager,Node,node,Cudd_NotPar(Res));
      node->state=CUDD_CALL_TREE_DONE;
      break;

    case CUDD_RETURN_ON_BOTH:
      if(Node>0) {
	CacheEntry=node->Then.CacheEntry;
	node->Then.Result=Res;
	node->state=CUDD_RETURN_ON_ELSE;		
      } else {
	CacheEntry=node->Else.CacheEntry;
	node->Else.Result=Res;
	node->state=CUDD_RETURN_ON_THEN;		
      }
      break;
    case CUDD_RETURN_ON_THEN:
      CacheEntry=node->Then.CacheEntry;
      node->Then.Result=Res;
      DdCallTreeCreateAndReturnRes(manager,abs(Node),node);
      node->state=CUDD_CALL_TREE_DONE;		
      break;
    case CUDD_RETURN_ON_ELSE:
      CacheEntry=node->Else.CacheEntry;
      node->Else.Result=Res;
      DdCallTreeCreateAndReturnRes(manager,abs(Node),node);
      node->state=CUDD_CALL_TREE_DONE;		
      break;
    case CUDD_OR_ON_BOTH:
      if(Node>0) {
	CacheEntry=node->Then.CacheEntry;
	node->Then.Result=Res;
	node->state=CUDD_OR_ON_ELSE;		
      } else {
	CacheEntry=node->Else.CacheEntry;
	node->Else.Result=Res;
	node->state=CUDD_OR_ON_THEN;		
      }
      break;
    case CUDD_OR_ON_THEN:
      CacheEntry=node->Then.CacheEntry;
      node->Then.Result=Res;
      DdCallTreeOr(manager,abs(Node));
      break;
  
    case CUDD_OR_ON_ELSE:
      CacheEntry=node->Else.CacheEntry;
      node->Else.Result=Res;
      DdCallTreeOr(manager,abs(Node));
      break;
    }

  /* updating cache with incomming result is common for all.
     This may also cause the other result to become available */
  if(CacheEntry!=NULL)
    cuddCacheUpdatePar(manager,CacheEntry,Res);

  
  return(1);
}

static int DdCallTreeOr(DdManager *manager,DdCallTreeNode Node)
{
  DdCallTreeNodeStruct *node;
  DdNodePar Then,Else;

  node=&manager->CallTree.NodeList[Node];
  Then=Cudd_NotPar(node->Then.Result);
  Else=Cudd_NotPar(node->Else.Result);
  node->state=CUDD_PROPAGATE_NEGATE;
  node->Then.CacheEntry=NULL;

  cuddBddRecurProcess2(manager,CUDD_BDD_AND,Then,Else,Node);
  return 1;
}
    

static int DdCallTreeReturnRes(DdManager *manager,DdCallTreeNode Node,
			       DdCallTreeNodeStruct *node,DdNodePar res)
{
  /* recursive update */
  if(node->ParentProcess==manager->Mypos)
    {
      if(DdCallTreeUpdateNode(manager,node->Parent,res)==0)
	res=CUDD_ERROR;
    }
  else
    {
      DdMessageReply(manager,node->ParentProcess,res,node->Parent);
    }  
  /* free node */
  node->Parent=manager->CallTree.FreeList;
  manager->CallTree.FreeList=abs(Node);

  return(!cuddEqualPar(res,CUDD_ERROR));
}

void DdCallTreeDestroy(DdManager *manager)
{
  FREE(manager->CallTree.NodeList);
}

int DdCallTreeResize(DdManager *manager)
{
  int i,oldSize=manager->CallTree.Size;
  DdCallTreeNodeStruct *newlist,*oldlist;
  printf("calltree %d\n",oldSize);fflush(stdout);
  oldlist=manager->CallTree.NodeList;
  manager->CallTree.Size=oldSize*DD_CALL_TREE_GROWTH;
  newlist=ALLOC(DdCallTreeNodeStruct,manager->CallTree.Size)-1; /* 1 indexed */
  if(newlist==NULL) 
    {
      manager->CallTree.Size=oldSize;
      assert(0);
      return(0);
    }
  for(i=1;i<=oldSize;i++)
    newlist[i]=oldlist[i];
  for(;i<manager->CallTree.Size;i++)
      newlist[i].Parent=i+1;
  newlist[manager->CallTree.Size].Parent=NULL_CALL_TREE_NODE;
  manager->CallTree.FreeList=oldSize+1;
  manager->CallTree.NodeList=newlist;
  free(oldlist+1); /* +1 because of the 1 index */ 
  manager->memused += 
    sizeof(DdCallTreeNodeStruct) * (manager->CallTree.Size-oldSize);
  return(1);
}

  
