#include "cuddInt.h"

/* could be made macro */
DdNodePar cuddNodeUserToPar(DdManager *manager,DdNode *node)
{
  if(Cudd_IsComplement(node))
    return(Cudd_NotPar(Cudd_Regular(node)->NodePar));
  else
    return(node->NodePar);
}

int cuddResizePageTableUser(DdManager *manager)
{
  int i,oldSize;
  DdNode **oldTable;
  DdUserNodeTable *mem;
  mem=&manager->UserNodes;

  oldSize=mem->TableSize;
  oldTable=mem->PageTable;
  mem->TableSize*=DD_PAGE_GROWTH; 
  mem->PageTable=ALLOC(DdNode *,mem->TableSize);
  if(mem->PageTable == NULL) {
    mem->PageTable=oldTable;
    mem->TableSize=oldSize;
    return(0);
  }

  for(i=0;i<oldSize;i++)
    mem->PageTable[i]=oldTable[i];

  manager->memused += sizeof(DdNode *) * (mem->TableSize-oldSize);
  return 1;
}


int cuddInitUserNodeTable(DdManager *manager)
{
  int j;
  DdNode **nodelist;

  manager->UserNodes.slots = CUDD_UNIQUE_SLOTS;
  manager->UserNodes.keys = 0;
  manager->UserNodes.maxKeys = CUDD_UNIQUE_SLOTS * DD_MAX_SUBTABLE_DENSITY;
  manager->UserNodes.dead = 0;

#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
  nodelist = manager->UserNodes.nodelist = ALLOC(DdNode *,CUDD_UNIQUE_SLOTS);
#ifdef __osf__
#pragma pointer_size restore
#endif
  if (nodelist == NULL) return(0);
  for (j = 0; (unsigned) j < CUDD_UNIQUE_SLOTS; j++) {
    nodelist[j] = NULL_USER;
  }

  manager->UserNodes.TableSize = DD_DEFAULT_PAGES;
  manager->UserNodes.PageTable = ALLOC(DdNode *,DD_DEFAULT_PAGES);
  if(manager->UserNodes.PageTable == NULL)
    return(0);
  manager->UserNodes.nextPage  = 0;
  manager->UserNodes.nextFree  = NULL_USER;

  manager->memused += sizeof(DdNode *) * CUDD_UNIQUE_SLOTS +
    sizeof(DdNode) * DD_DEFAULT_PAGES;

  return(1);
}

DdNode* cuddUserNodeInter(DdManager *manager,DdNodePar node_par)
{
  int pos;
  DdNode **nodelist;
  DdNode *looking;
  int complement;

  if(cuddEqualPar(node_par,CUDD_ERROR))
    return(NULL);

  if (manager->UserNodes.keys > manager->UserNodes.maxKeys) {
    cuddUserNodeRehash(manager);
  }
  
  if((complement=Cudd_IsComplementPar(node_par)))
    node_par=Cudd_NotPar(node_par);
  
  pos = ddHashUser(node_par,manager->UserNodes.slots);
  nodelist = manager->UserNodes.nodelist;
  looking = nodelist[pos];
    
  while (!cuddEqual(looking,NULL_USER)) {
    if (cuddEqualPar(looking->NodePar,node_par) )
      {
	return(Cudd_NotCond(looking,complement));  
      }
    looking = looking->next;
  }

  manager->UserNodes.keys++;

  looking = cuddAllocNodeUser(manager);
  if (cuddEqual(looking,NULL_USER)) return(NULL_USER);
  looking->index=manager->InvOrigPerm[node_par.index];
  looking->ref = 0;
  looking->NodePar = node_par;
  looking->next = nodelist[pos];
  nodelist[pos] = looking;

  return(Cudd_NotCond(looking,complement)); 
}

void cuddComputeNewOrigPerm(DdManager *manager)
{
  int var;
  for(var=0;var<manager->size;var++)
    {
      manager->OrigPerm[var]=(manager->vars[var])->NodePar.index;
      manager->InvOrigPerm[(manager->vars[var])->NodePar.index]=var;
    }
}

DdNode* cuddAllocNodeUser(DdManager *manager)
{
    int i;
    DdNode *page,*node;
    int pageNumber;
    

    if (cuddEqual(manager->UserNodes.nextFree,NULL_USER)) 
      { /* memory is all full */
	if(manager->UserNodes.nextPage==NULL_PAGE)
	  return(NULL_USER); /* No more pages available */ 

	if(manager->UserNodes.nextPage==manager->UserNodes.TableSize)
	  if(cuddResizePageTableUser(manager)==0)
	    return(NULL_USER); /* unable to allocate mem for new entries */

	pageNumber=manager->UserNodes.nextPage;
	page = manager->UserNodes.PageTable[pageNumber] = ALLOC(DdNode,DD_MEM_CHUNK) - 1 ; 
	/* -1 to make it 1 indexed */ 
#ifdef __osf__
#pragma pointer_size restore
#endif
	if (page == NULL) {
#ifdef DD_VERBOSE
	  (void) fprintf(stderr,"cuddAllocNode: out of memory");
	  (void) fprintf(stderr,"Memory in use = %ld\n", manager->memused);
#endif
	  return(NULL_USER);
	}
	manager->memused += (DD_MEM_CHUNK) * sizeof(DdNode);

	for(i=1;i<DD_MEM_CHUNK;i++)
	  {
	    page[i].next=&page[i+1];
	  } 

	page[DD_MEM_CHUNK].next = NULL_USER;
	
	manager->UserNodes.nextFree = &page[1];
	manager->UserNodes.nextPage++;
      }

    node = manager->UserNodes.nextFree;
    manager->UserNodes.nextFree = node->next;

    return(node);

} /* end of cuddAllocNodeUser */

int cuddDeallocNodeUser(DdManager *manager,DdNode *node)
{
  assert(node->ref==0);
  node->next=manager->UserNodes.nextFree;
  manager->UserNodes.nextFree=node;
  manager->UserNodes.dead++;
}

int cuddUserNodeRehash(DdManager *manager)
{
    unsigned int slots, oldslots;
    int j,pos;
    DdNode **nodelist, **oldnodelist;
    DdNode *node;
    DdNode *next;

    oldslots = manager->UserNodes.slots;
    oldnodelist = manager->UserNodes.nodelist;

    /* Compute the new size of the subtable */
    slots = oldslots;    
    while(manager->UserNodes.keys > slots * DD_MAX_SUBTABLE_DENSITY)
      slots = Cudd_Prime(DD_GROW_FACTOR * slots);

    manager->UserNodes.slots = slots;
    manager->UserNodes.maxKeys = slots * DD_MAX_SUBTABLE_DENSITY;

#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
    nodelist = manager->UserNodes.nodelist = ALLOC(DdNode *, slots);
#ifdef __osf__
#pragma pointer_size restore
#endif

    /* init */
    for (j = 0; (unsigned) j < slots; j++) {
      nodelist[j] = NULL_USER;
    }

    /* rehash */
    for (j = 0; (unsigned) j < oldslots; j++) {
      node = oldnodelist[j];
      while (!cuddEqual(node, NULL_USER)) 
	{
	  next = node->next;
	  pos = ddHashUser(node->NodePar,slots);
	  node->next = nodelist[pos];
	  nodelist[pos] = node;
	  node = next;
	}
    }  
    FREE(oldnodelist);

#ifdef DD_VERBOSE
    (void) fprintf(stderr,
		   "rehashing user nodes: keys %d dead %d new size %d\n",
		   manager->UserNodes.keys,manager->UserNodes.dead,slots);
#endif

    /* Update global data */
    manager->memused += (slots - oldslots) * sizeof(DdNode *);

} /* end of cuddUserNodeRehash */

int cuddUserNodeGarbageCollect(DdManager *manager)
{
  DdNode *node, **last;
  int i;

  for (i = 0; i < manager->UserNodes.slots; i++) 
    {
      last = &manager->UserNodes.nodelist[i];
      node = manager->UserNodes.nodelist[i];
      while (node!=NULL)
	{
	  if(node->ref==0)
	    {
	      *last=node->next;
	      cuddDeallocNodeUser(manager,node);
	      node=*last;
	    }	    
	  else
	    { 
	      last=&node->next;
	      node=node->next;
	    }
	}
    }
  manager->UserNodes.keys-=manager->UserNodes.dead;
  manager->UserNodes.dead=0;
  return(1);
}

int cuddUpdateUserNodes(DdManager *manager)
{
  int i;
  DdNode *node;

  /* update user nodes wityh news location od par node */
  for (i = 0; i < manager->UserNodes.slots; i++) 
    {
      node = manager->UserNodes.nodelist[i];
      while (node!=NULL)
	{
	  if(node->index<CUDD_MAXINDEX)
	    node->NodePar=manager->memory[node->NodePar.index].PageTable2[node->NodePar.address.page][node->NodePar.address.number].NewLocation;
	  node=node->next;
	}
    }

  cuddUserNodeRehash(manager);
  cuddComputeNewOrigPerm(manager);

  /* update index of user nodes */
  for (i = 0; i < manager->UserNodes.slots; i++) 
    {
      node = manager->UserNodes.nodelist[i];
      while (node!=NULL)
	{
	  if(node->index<CUDD_MAXINDEX)
	    node->index=manager->InvOrigPerm[node->NodePar.index];
	  node=node->next;
	}
    }
  return(1);
}

