#ifndef _MESSAGE
#define _MESSAGE

#include "cuddInt.h"
#include "pvm3.h"

#ifdef USE_PGTRACE
#include "pgpvm2.h"
#endif 

#define CUDD_SLAVE_PROGRAM "process"

#define NONE_RECIEVED 0
#define THEN_RECIEVED 1
#define ELSE_RECIEVED 2
#define BOTH_RECIEVED 3

/* note theese constants must not overlap with the operation constanst
   (CUDD_BDD_AND etc.) */
#define CUDD_NO_MORE_MESSAGES 100
#define CUDD_END              101
#define CUDD_BDD_REPLY        102
#define CUDD_NEW_PROCESS      103
#define CUDD_FLUSH_DOWN       104
#define CUDD_FLUSH_UP         105
#define CUDD_NEW_VAR          106

#define CUDD_SEND_INDEX       107
#define CUDD_RECIEVE_INDEX    108
#define CUDD_RETRIEVE_INDEX   109
#define CUDD_REDISTRIBUTE_LEVELS   112

#define CUDD_REORDER          110
#define CUDD_DUMP_DOT         111
#define CUDD_FORWARD_OPERATION 113
#define CUDD_FORWARD_REPLY     114
#define CUDD_FORCE_REPLY       115

#define DD_DOWN_MESSAGE 0
#define DD_UP_MESSAGE 1

#define MAX_PROC 20

#define DD_MAX_OPS               8
#define DD_BUFFER_SIZE        1000
#define DD_DOWN_BUFFER_SIZE      DD_BUFFER_SIZE
#define DD_UP_BUFFER_SIZE        DD_BUFFER_SIZE
#define DD_DOWN_BUFFER_GROWTH    1
#define DD_UP_BUFFER_GROWTH      1
#define DD_DOWN_BUFFER_MAX  100000
#define DD_UP_BUFFER_MAX    100000

#define DD_NODES_BETWEEN_REDISTRUBUTION 100000

#define NULL_MESSAGE_ID -1

typedef int ProcessId;
typedef int MessageId;

typedef struct Message
{
  int Type;
  DdNodePar f,g,h;
  ProcessId Sender;
  ProcessId Reciever;
  MessageId Id;
} Message;

typedef struct
{
  int n;
  int n_per_op[DD_MAX_OPS];
  struct {
    int from;
    int op;
    int to;
  } ForwardInfo[DD_MAX_OPS];
  int n_op;
  int ops[DD_MAX_OPS];
  DdHalfWord       f_index [DD_MAX_OPS][DD_BUFFER_SIZE];
  DdHalfWord       f_page  [DD_MAX_OPS][DD_BUFFER_SIZE];
  DdSignedHalfWord f_number[DD_MAX_OPS][DD_BUFFER_SIZE];
  DdHalfWord       g_index [DD_MAX_OPS][DD_BUFFER_SIZE];
  DdHalfWord       g_page  [DD_MAX_OPS][DD_BUFFER_SIZE];
  DdSignedHalfWord g_number[DD_MAX_OPS][DD_BUFFER_SIZE];
  DdCallTreeNode   caller  [DD_MAX_OPS][DD_BUFFER_SIZE];
} message_buffer;

typedef struct ProcessInfo
{
  ProcessId Parent;
  ProcessId Tids[MAX_PROC];
  unsigned int Nprocess;
  ProcessId Mytid;
  int Nnode;
  struct pvmhostinfo *hostp;     /* machine information 
				    (only present at master) */
  int flush_count;
  int flush;
  int down_flag;                 /* indication if any messages have been
				    send down the pipeline */
  int Encoding;                  /* the encoding for pvm */

  int DownCount;                 /* current # messages in buffer */
  message_buffer DownBuffer;
  int DownBuf;                    /* bufid */ 
  int DownCountSend;             /* # messages collected before sending */
  int DownSendMax;               /* Do not increase buffer beyond this */ 
  int UpCount;                   /* current # messages in buffer */
  message_buffer UpBuffer;
  int UpBuf;                      /* bufid */ 
  int UpCountSend;               /* # messages collected before sending */
  int UpSendMax;                 /* Do not increase buffer beyond this */ 

  int NextRedistribution;

#ifdef DD_COMM_MONITOR
  int DownMessages;
  int UpMessages;
  int Forwards;
  int ITE;
  int WorkDown;
#endif
} ProcessInfo;


/******************************* MACRO ******************************/
#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define PVM_pkDdHalfWord  pvm_pkuint
#define PVM_upkDdHalfWord pvm_upkuint

#define PVM_pkDdSignedHalfWord  pvm_pkint
#define PVM_upkDdSignedHalfWord pvm_upkint

#else
#define PVM_pkDdHalfWord  pvm_pkushort
#define PVM_upkDdHalfWord pvm_upkushort

#define PVM_pkDdSignedHalfWord  pvm_pkshort
#define PVM_upkDdSignedHalfWord pvm_upkshort
#endif

#define PVM_pkDdNodePar(f) (PVM_pkDdHalfWord(&f.index,1,1)>=0 && PVM_pkDdHalfWord(&f.address.page,1,1)>=0 && PVM_pkDdSignedHalfWord(&f.address.number,1,1)>=0)
#define PVM_upkDdNodePar(f) ( PVM_upkDdHalfWord(&f.index,1,1)>=0 &&   PVM_upkDdHalfWord(&f.address.page,1,1)>=0 &&   PVM_upkDdSignedHalfWord(&f.address.number,1,1)>=0)

#define PVM_pkDdNodeAdr(a)  ( PVM_pkDdHalfWord(&a.page,1,1)>=0 && PVM_pkDdSignedHalfWord(&a.number,1,1)>=0)
#define PVM_upkDdNodeAdr(a)  ( PVM_upkDdHalfWord(&a.page,1,1)>=0 && PVM_upkDdSignedHalfWord(&a.number,1,1)>=0)

#define PVM_pkDdCallTreeNode(node) pvm_pkint(&node,1,1)
#define PVM_upkDdCallTreeNode(node) pvm_upkint(&node,1,1)

/************* message data encaptiolation */
#define messageSendIte(man,f,g,h)   DdMessageSendMaster3(man,CUDD_BDD_ITE,f,g,h) 
#define messageSendAndAbstract(man,f,g,cube)     \
DdMessageSendMaster3(man,CUDD_BDD_ANDABSTRACT,f,g,cube)
#define messageSendAnd(man,f,g)     DdMessageSendMaster2(man,CUDD_BDD_AND,f,g)
#define messageSendXor(man,f,g)     DdMessageSendMaster2(man,CUDD_BDD_XOR,f,g)
#define messageSendExists(man,f,g)  DdMessageSendMaster2(man,CUDD_BDD_EXISTS,f,g)
#define messageSendIntersect(man,f,g) assert(0 && (int)"intersect not implemented");

/******************* interface to PVM */

/* PVM_interface.c */
extern int DdMessageSend2(DdManager *manager,int op, int index, int from,
			  DdNodePar f,DdNodePar g,DdCallTreeNode caller);
extern int DdMessageReply(DdManager *manager,int to,DdNodePar res,DdCallTreeNode caller);
extern int flush_buffer_down(DdManager *manager);
extern int flush_buffer_up(DdManager *manager);
int cuddSendExternRef(DdManager *manager);
int cuddExternGetNewLocations(DdManager *manager,DdMemory *old_mems) ;

/* PVM_master.c */
extern DdNodePar DdAssignNewVar(DdManager *manager);
extern DdNodePar DdMessageRecieveMaster(DdManager *manager);
extern int DdInitMaster(DdManager *manager);
extern int DdMessageSendMaster2(DdManager *manager,int op,DdNodePar f,DdNodePar g);
extern int DdMessageSendMaster3(DdManager *manager,int op,DdNodePar f,DdNodePar g,DdNodePar h);
extern int InitProcesses(DdManager *manager,int cache_size,int cache_size_max);
extern int KillProcesses(DdManager *manager);

/* PVM_slave.c */
int cuddRedistributeLevelsSlave(DdManager *manager);
int cuddSendProfile(DdManager *manager);
int cuddReorderUpdateTableSizesSlave(DdManager *manager);
int cuddGetExternRef(DdManager *manager);
int cuddExternDistributeNewLocations(DdManager *manager,DdMemory *old_mems);
extern int DdMessagesRecieve(DdManager *manager);
int DdReplyRecieve(DdManager *manager);
extern int DdInitSlave(DdManager *manager,int *cache_size, int *cache_size_max);
int cuddDumpDotSlave(DdManager *manager,char file_name[]);

#define cuddStoreForSending(manager,f) (void)PVM_pkDdNodePar(f)
#define cuddInitExternDistributeNewLocations(manager,old_mems) pvm_setsbuf(manager->ProcessInfo->UpBuf);

#endif

